function RCO(func,D0,tMax) % Ruler and Compass Optimisation
%{
    Maurice.Clerc@WriteMe.com
   
A simple deterministic optimisation method.
On dimension 1 you just need a ruler and a compass to apply it.
And also on higher dimension D ... assuming you have
a D-ruler and a D-compass!

2024-05-1 version
WARNING: This does not always match the online papers. 
         Please check the dates.

Main conclusions:
- Only for low dimension (typically <10)
- Cant't find a solution on the boundary. But you can "artificially"
  extend the definition space thanks to fMax (see function Eggholder +)
- Doesn't work well on (partly) discrete problems
- Can be pretty good on some functions supposed to be difficult
- May stagnate in a local minimum if its value is very close to
  the one of the global minimum.

Examples
RCO(4,1,10) % Bi-parabola
RCO(10,2,300) % Six Hump Camel Back
RCO(9,3,5000) %  Rosenbrock
RCO(1,4,10000) % Rastrigin
RCO(49,10,50000) % Lévy

%}

warning('off') % Don't display warnings (in particular for singular matrix)
rng(123456789) % For reproducibility of test problems that use rand
global noWayCount % Number of times no intersection can be computed
global outsideCount % Number of time the intersection
                    % is outside the search space
global t % Number of steps
t=0;
noWayCount=zeros(1,tMax); 
outsideCount=zeros(1,tMax); 
global exploit
exploit=zeros(1,tMax);
global fMax

adaptLBcoeff=0.5; %0.5?; % In ]0,1[
adaptOption=0; % 0, 1, 2,  3, 4  
                %{ 
                    Adaptation method. It may be useful to try all.
                    Note: with adaptLBcoeff=0.5 
                     adaptOption 1 and 3 are the same 
                    If adaptLBcoeff>0.5 then 3 seems better
                %}
stagDetect=false; % If true: stagnation detection is used and
                    %  and a new point is defined just by mowing
                    % 'towards' a corner.
global x0 % Set of corners
global toCorner % Number of times a corner is used
                % to construct a new point (except at the very beginning)
toCorner=0;

% Possible plots for D=1 or 2
plotLandscape=true;
plotLine=true;
plotPoints=true;

% Possible plots for any problem
plotNowayOutside=false;
plotBalance=false;

fprintf('\n=============================================================')
% Define the problem
[xMin, xMax, lowerBound,quantis,D]=get_func(func,D0);

fprintf('\n function %i, dimension %i, budget %i',func,D,tMax);

tStart=cputime; % Start for computing time
% -------------------------------- Initialisation
% Define all "corners" (2^D)
x=corners(xMin,xMax);
x0=x; % May be useful for out2in

% Evaluate them
[nCorners,~]=size(x);
for n=1:nCorners
  f(n)=fit(x(n,:),func);
end
fMax=max(f); % Useful to "extend" a definition space

% Save the best
[fBest,Ind]=min(f);
fprintf('\n Best after init %e',fBest)
xBest=x(Ind,:);
fBestQ=fBest; % For (partly) discrete problems

plotRun=(plotPoints || plotLine || plotDash) && D<=2;
if plotRun
  for n=1:nCorners
    line(n,1:D)=x(n,:); line(n,D+1)=f(n);
  end
end

% --------------------------------
noImprov=0; % Number of improvements

while t<tMax
  %fprintf('\n t %i, fBest %e',t,fBest)
    if ~isempty(quantis)
      fB=fBestQ;
    else
      fB=fBest;
    end
    
    switch adaptOption
      case 1 % Method 1
        if fB>=0
          lb=adaptLBcoeff*fB;
        else
          lb=(1+adaptLBcoeff)*fB;
        end
        
      case 2 % Method 2
        lb=adaptLBcoeff*lowerBound +(1-adaptLBcoeff)*fB;
        
      case 3 % Method 3
        if fB>0
          lb=fB*adaptLBcoeff;
        else
            lb=fB*(2-adaptLBcoeff);
        end
        
      case 4
        lb=fB;
    %end
      otherwise % User defined lower bound
    lb=lowerBound;
    end
  LB(t)=lb; % Useful for plot
  %fprintf('\n t %i, lb %e',t,lb)
 
 if stagDetect
  if noImprov>nCorners  % If no improvement during a cycle
    % try another method in case of intersection
    % outside the search space
    out2inOption=1;
    noImprov=0;
  else
    out2inOption=0;
  end
 else
  out2inOption=0;
 end
 
  % Define a new position
  [xn,fn]=new(x,f,xMin,xMax,func,lb,out2inOption);
  %fprintf('\n RCO 155');disp(t)
  if plotRun % Save positions for future plots
    line(t,1:D)=xn;
    line(t,D+1)=fn;
  end
  
  % For (partly) discrete problems
  if ~isempty(quantis)
    if fn<fBestQ % Keep the best solution after discretisation
      xQ=quantify(xn,quantis);
      fQ=fit(xQ,func);
      if fQ<fBestQ
        fBestQ=fQ;
        xBest=xQ;
      else
        noImprov=noImprov+1;
      end
    end
  else
    if fn<fBest
      fBest=fn;
      xBest=xn;
    else
      noImprov=noImprov+1;
    end
  end
  
  % Define another line (D=1) or plane (D=2) or hyperplane (D>2)
  % (treadmill method)
  for p=1:nCorners-1
    x(p,:)=x(p+1,:);
    f(p)=f(p+1);
  end
  x(nCorners,:)=xn;
  f(nCorners)=fn;
end % end of the run

tEnd=cputime; % Stop computing time
% -------------------------------- Results
if ~isempty(quantis)  % Some problems may be (partly) discrete
    fBest=fBestQ;
end

if stagDetect
  fprintf('\n Stagnation detection activated');
end
  switch adaptOption
    case 1
      fprintf('\n Adaptation 1')
      fprintf('\n Totally adaptive lower bound, coefficient %f',adaptLBcoeff);
    case 2
      fprintf('\n Adaptation 2');
      fprintf('\n Mix user defined lower bound %e + adaptive,  coefficient %f',...
        lowerBound,adaptLBcoeff);
      fprintf('\n (no matter the sign of fBest)')
    case 3
      fprintf('\n Adaptation 3');
       fprintf('\n Totally adaptive lower bound, coefficient %f',adaptLBcoeff);

    case 4
      fprintf('\n Adaptation 4, Lowerbound=fBest'); % BAD
  otherwise
  fprintf('\n lowerBound %f',lowerBound);
  end

fprintf('\n Evaluations %i', t);
fprintf('\n noWayCount %i',sum(noWayCount));
fprintf('\n outsideCount %i',-sum(outsideCount));
fprintf('\n toCorner %i',toCorner);
fprintf('\n final fMax %e',fMax)

fprintf('\n fBest %e',fBest)
fprintf('\n at \n'); disp(xBest)

fprintf('\n CPU time %e \n',tEnd-tStart)

% ================================== Plots
if plotRun 
  LB=zeros(tMax,D+1); % Just to save the lower bounds.
  figure; 
end

if plotLandscape && D<=2
  fPlot(func,D,xMin,xMax,lowerBound);
  hold on
end

if plotLine
  switch D
    case 1
      plot(line(:,1), line(:,2),'r-','LineWidth',2);
      hold on
    case 2
      plot3(line(:,1), line(:,2),line(:,3),'r-','LineWidth',2);
      hold on
  end
end

if plotPoints
  switch D
    case 1
      plot(line(:,1), line(:,2),'r.','MarkerSize',12)
      %axis([xMin,xMax,lowerBound, max(line(:,2))]);
      hold on
    case 2
      plot3(line(:,1), line(:,2),line(:,3),'r.','MarkerSize',12)
      hold on
  end
end


if plotNowayOutside
  figure
  bar(noWayCount,0.1, 'k');
  hold on
  bar(outsideCount,0.1,'k');
  xlabel("Evaluations");
  if sum(noWayCount)>0
   label_h = ylabel('outside | noway');
  else
    label_h = ylabel('outside');
  end
   % label_h.Position(2) = 0; % change vertical position of ylabel
    set(gca, 'YTick', []);  
end

if plotBalance % Normally associated to plotNowayOutside
   balance(1)=0; % Exploration/exploitation
   for fe=2:tMax
     exploit(fe)=exploit(fe-1)+ exploit(fe);
     balance(fe)=(fe-exploit(fe))/exploit(fe);
   end
   if ~plotNowayOutside
     figure
   else
     yyaxis right
   end
   plot(balance(2^D:tMax),"Color","r", 'LineWidth',2)
   xlabel("Evaluations");
   %yyaxis right
   ylabel("Exploration/exploitation", "Color", "r")
end
end
 



